
sub salsa20_stream {
    my ($args) = @_;
    delete $args->{'data'};
    my $salsa20 = salsa20($args);
    my $buffer = '';
    return sub {
        my $enc = shift;
        $buffer .= $salsa20->("\0" x 64) while length($buffer) < length($enc);
        my $data = join '', map {chr(ord(substr $enc, $_, 1) ^ ord(substr $buffer, $_, 1))} 0 .. length($enc)-1;
        substr $buffer, 0, length($enc), '';
        return $data;
    };
}

sub salsa20 { # http://cr.yp.to/snuffle/salsa20/regs/salsa20.c
    my ($args) = @_;
    my ($key, $iv, $rounds) = @$args{qw(key iv rounds)};
    $rounds ||= 20;

    my (@k, @c);
    if (32 == length $key) {
        @k = unpack 'L8', $key;
        @c = (0x61707865, 0x3320646e, 0x79622d32, 0x6b206574); # SIGMA
    } elsif (16 == length $key) {
        @k = unpack 'L8', $key x 2;
        @c = (0x61707865, 0x3120646e, 0x79622d36, 0x6b206574); # TAU
    } else {
        die "Salsa20 key length must be 16 or 32\n";
    }
    die "Salsa20 IV length must be 8\n" if length($iv) != 8;
    die "Salsa20 rounds must be 8, 12, or 20.\n" if !grep {$rounds != $_} 8, 12, 20;
    my @v = unpack('L2', $iv);

    #            0                                  5      6      7            10                                 # 15
    my @state = ($c[0], $k[0], $k[1], $k[2], $k[3], $c[1], $v[0], $v[1], 0, 0, $c[2], $k[4], $k[5], $k[6], $k[7], $c[3]);

    my $rotl32 = sub { return (($_[0] << $_[1]) | ($_[0] >> (32 - $_[1]))) % 0x100000000 };
    my $word_to_byte = sub {
        my @x = @state;
        for (1 .. $rounds/2) {
            $x[ 4] ^= $rotl32->(($x[ 0] + $x[12]) % 0x100000000,  7);
            $x[ 8] ^= $rotl32->(($x[ 4] + $x[ 0]) % 0x100000000,  9);
            $x[12] ^= $rotl32->(($x[ 8] + $x[ 4]) % 0x100000000, 13);
            $x[ 0] ^= $rotl32->(($x[12] + $x[ 8]) % 0x100000000, 18);
            $x[ 9] ^= $rotl32->(($x[ 5] + $x[ 1]) % 0x100000000,  7);
            $x[13] ^= $rotl32->(($x[ 9] + $x[ 5]) % 0x100000000,  9);
            $x[ 1] ^= $rotl32->(($x[13] + $x[ 9]) % 0x100000000, 13);
            $x[ 5] ^= $rotl32->(($x[ 1] + $x[13]) % 0x100000000, 18);
            $x[14] ^= $rotl32->(($x[10] + $x[ 6]) % 0x100000000,  7);
            $x[ 2] ^= $rotl32->(($x[14] + $x[10]) % 0x100000000,  9);
            $x[ 6] ^= $rotl32->(($x[ 2] + $x[14]) % 0x100000000, 13);
            $x[10] ^= $rotl32->(($x[ 6] + $x[ 2]) % 0x100000000, 18);
            $x[ 3] ^= $rotl32->(($x[15] + $x[11]) % 0x100000000,  7);
            $x[ 7] ^= $rotl32->(($x[ 3] + $x[15]) % 0x100000000,  9);
            $x[11] ^= $rotl32->(($x[ 7] + $x[ 3]) % 0x100000000, 13);
            $x[15] ^= $rotl32->(($x[11] + $x[ 7]) % 0x100000000, 18);

            $x[ 1] ^= $rotl32->(($x[ 0] + $x[ 3]) % 0x100000000,  7);
            $x[ 2] ^= $rotl32->(($x[ 1] + $x[ 0]) % 0x100000000,  9);
            $x[ 3] ^= $rotl32->(($x[ 2] + $x[ 1]) % 0x100000000, 13);
            $x[ 0] ^= $rotl32->(($x[ 3] + $x[ 2]) % 0x100000000, 18);
            $x[ 6] ^= $rotl32->(($x[ 5] + $x[ 4]) % 0x100000000,  7);
            $x[ 7] ^= $rotl32->(($x[ 6] + $x[ 5]) % 0x100000000,  9);
            $x[ 4] ^= $rotl32->(($x[ 7] + $x[ 6]) % 0x100000000, 13);
            $x[ 5] ^= $rotl32->(($x[ 4] + $x[ 7]) % 0x100000000, 18);
            $x[11] ^= $rotl32->(($x[10] + $x[ 9]) % 0x100000000,  7);
            $x[ 8] ^= $rotl32->(($x[11] + $x[10]) % 0x100000000,  9);
            $x[ 9] ^= $rotl32->(($x[ 8] + $x[11]) % 0x100000000, 13);
            $x[10] ^= $rotl32->(($x[ 9] + $x[ 8]) % 0x100000000, 18);
            $x[12] ^= $rotl32->(($x[15] + $x[14]) % 0x100000000,  7);
            $x[13] ^= $rotl32->(($x[12] + $x[15]) % 0x100000000,  9);
            $x[14] ^= $rotl32->(($x[13] + $x[12]) % 0x100000000, 13);
            $x[15] ^= $rotl32->(($x[14] + $x[13]) % 0x100000000, 18);
        }
        return pack 'L16', map {($x[$_] + $state[$_]) % 0x100000000} 0 .. 15;
    };

    my $encoder = sub {
        my $enc = shift;
        my $out = '';
        while (length $enc) {
            my $stream = $word_to_byte->();
            $state[8] = ($state[8] + 1) % 0x100000000;
            $state[9] = ($state[9] + 1) % 0x100000000 if $state[8] == 0;
            my $chunk = substr $enc, 0, 64, '';
            $out .= join '', map {chr(ord(substr $stream, $_, 1) ^ ord(substr $chunk, $_, 1))} 0 .. length($chunk)-1;
        }
        return $out;
    };
    return $encoder if !exists $args->{'data'};
    return $encoder->(defined($args->{'data'}) ? $args->{'data'} : '');
}

sub dump_stream
{
	my ($data)=@_;
	my $i = 0;
    	foreach (unpack( 'H32' x (length( $data )/16), $data )) {
		printf "%30s %s", $i%4 ? "" : " Stream[" . (64*$i) . ".." . (64*($i+1)) . "] = " , uc "$_\n"; 
		$i++;
	}
}
sub dump_key
{
	my ($data)=@_;
	my $i = 0;
    	foreach (unpack( 'H32' x (length( $data )/16), $data )) {
		printf "%30s %s", $i%4 ? "" : " key = " , uc "$_\n"; 
		$i++;
	}
}
sub dump_iv
{
	my ($data)=@_;
	printf "%30s %s", " IV = " , uc unpack ('H*', $data) ."\n"; 
}
my @salsa20_testvectors = (
		{ key => "8000000000000000000000000000000000000000000000000000000000000000", IV => "0000000000000000" , set => 1, vector => 0},
		{ key => "2425262728292A2B2C2D2E2F303132333435363738393A3B3C3D3E3F40414243", IV => "0000000000000000" , set => 3, vector => 36},
		{ key => "0558ABFE51A4F74A9DF04396E93C8FE23588DB2E81D4277ACD2073C6196CBF12", IV => "167DE44BB21980E7" , set => 6, vector => 1},
		);
my $input = "\0" x 512;
$\ = "\n";
print "Cleartext input for these tests is a 512-byte block of zeros, just like http://www.ecrypt.eu.org/stream/svn/viewcvs.cgi/ecrypt/trunk/submissions/salsa20/full/verified.test-vectors?logsort=rev&rev=210&view=markup";

foreach (@salsa20_testvectors) {
	my ($key, $IV) = (pack("H*", $_->{key}), pack("H*", $_->{IV}));
     	my $encoder = salsa20_stream({key => $key, iv => $IV}); # no data
	my $encoded = $encoder->($input); 

	print "Set $_->{set}, vector # $_->{vector}:";
	dump_key($key);
	dump_iv($IV);
	dump_stream($encoded);
}
